/**
 * @file libgalago/galago-dbus.c D-BUS Support API
 *
 * @Copyright (C) 2004-2005 Christian Hammond.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
#include <libgalago/galago-dbus.h>
#include <libgalago/galago-assert.h>
#include <libgalago/galago-status.h>

void
galago_dbus_message_iter_append_string_or_nil(DBusMessageIter *iter,
											  const char *str)
{
	galago_return_if_fail(iter != NULL);

	dbus_message_iter_append_string(iter, (str == NULL ? "" : str));
}

char *
galago_dbus_message_iter_get_string_or_nil(DBusMessageIter *iter)
{
	char *str;

	galago_return_val_if_fail(iter != NULL, NULL);

	str = dbus_message_iter_get_string(iter);

	if (str == NULL || *str == '\0')
		str = NULL;

	return str;
}

void
galago_dbus_message_iter_append_object(DBusMessageIter *iter,
									   const void *object)
{
	galago_return_if_fail(iter != NULL);
	galago_return_if_fail(object == NULL || GALAGO_IS_OBJECT(object));

	if (object == NULL)
		dbus_message_iter_append_nil(iter);
	else
	{
		GalagoObjectClass *klass;

		klass = GALAGO_OBJECT_CLASS(object);

		if (klass->dbus_message_append != NULL)
			klass->dbus_message_append(iter, object);
		else
		{
			galago_log_error("Class type %s passed to "
							 "galago_dbus_message_iter_append_object does not "
							 "implement dbus_message_append!\n",
							 galago_class_get_name(klass));
		}
	}
}

void
galago_dbus_message_iter_append_object_list(DBusMessageIter *iter,
											const GalagoList *list)
{
	const GalagoList *l;
	size_t num_objects;

	galago_return_if_fail(iter != NULL);

	num_objects = galago_list_get_count(list);

	dbus_message_iter_append_uint32(iter, num_objects);

	/*
	 * NOTE: There's a possible problem of objects in a list being
	 *       of different classes. If this is the case, oops, screwed.
	 *       Up to the developer to fix.
	 */
	for (l = list; l != NULL; l = l->next)
	{
		GalagoObject *object = (GalagoObject *)l->data;

		galago_dbus_message_iter_append_object(iter, object);
	}
}

void *
galago_dbus_message_iter_get_object(DBusMessageIter *iter,
									const GalagoObjectClass *klass)
{
	GalagoObject *object;

	galago_return_val_if_fail(iter  != NULL, NULL);
	galago_return_val_if_fail(klass != NULL, NULL);

	if (dbus_message_iter_get_arg_type(iter) == DBUS_TYPE_NIL)
		return NULL;

	if (klass->dbus_message_get == NULL)
	{
		galago_log_error("Class type %s passed to "
						 "galago_dbus_message_iter_get_object does not "
						 "implement dbus_message_get!\n",
						 galago_class_get_name(klass));
		return NULL;
	}

	object = klass->dbus_message_get(iter);

	return object;
}

GalagoList *
galago_dbus_message_iter_get_object_list(DBusMessageIter *iter,
										 const GalagoObjectClass *klass)
{
	GalagoList *list = NULL;
	dbus_uint32_t num_objects, i;

	galago_return_val_if_fail(iter  != NULL, NULL);
	galago_return_val_if_fail(klass != NULL, NULL);

	num_objects = dbus_message_iter_get_uint32(iter);

	for (i = 0; i < num_objects; i++)
	{
		GalagoObject *object;

		dbus_message_iter_next(iter);

		object = galago_dbus_message_iter_get_object(iter, klass);

		list = galago_list_append(list, object);
	}

	return list;
}

DBusMessage *
galago_dbus_message_new_method_call(const void *object, const char *name,
									galago_bool reply, DBusMessageIter *iter)
{
	DBusMessage *message;
	GalagoObjectClass *klass;
	const char *obj_path;
	const char *iface;

	galago_return_val_if_fail(object != NULL, NULL);
	galago_return_val_if_fail(name   != NULL, NULL);
	galago_return_val_if_fail(*name  != '\0', NULL);
	galago_return_val_if_fail(GALAGO_IS_OBJECT(object), NULL);

	klass = GALAGO_OBJECT_CLASS(object);

	obj_path = galago_object_get_dbus_path(object);
	iface    = galago_class_get_dbus_iface(klass);

	if (obj_path == NULL)
	{
		galago_log_error(
			"No object path was registered for class '%s'. "
			"Please report this.\n",
			galago_class_get_name(klass));

		return NULL;
	}

	if (iface == NULL)
	{
		galago_log_error(
			"No D-BUS interface was registered for class '%s'. "
			"Please report this.\n",
			galago_class_get_name(klass));

		return NULL;
	}

	message = dbus_message_new_method_call(GALAGO_DBUS_SERVICE,
										   obj_path, iface, name);

	galago_return_val_if_fail(message != NULL, NULL);

	dbus_message_set_no_reply(message, !reply);

	if (iter != NULL)
		dbus_message_iter_init(message, iter);

	return message;
}

void
galago_dbus_object_push_full(void *object)
{
	GalagoObjectClass *klass;

	galago_return_if_fail(object != NULL);
	galago_return_if_fail(GALAGO_IS_OBJECT(object));

	klass = GALAGO_OBJECT_CLASS(object);

	if (klass->dbus_push_full != NULL)
		klass->dbus_push_full(object);
	else
	{
		galago_log_error("Class type %s passed to "
						 "galago_dbus_object_push_full does not "
						 "implement dbus_push_full!\n",
						 galago_class_get_name(klass));
	}
}
